require 'rest-client'
require 'openssl'
require 'base64'
require 'json'
#require 'URI'
require 'stringio'

# locally:
# cp LuxSci.rb /usr/share/ruby/2.0.0/

module LuxSci

class API
  
  @@username = ''
  @@password = ''

  @@apiKey = nil
  @@apiSecret = nil
  @@authCode = nil

  @@in_login = false
  @@auto_auth = false

  @@baseURI = nil
  @@debug = false

  @@sslVerify = true

  def self.sslVerify
    @@sslVerify
  end
  def self.sslVerify=(torf)
    @@sslVerify=torf
  end

  def self.userName=(username)
    @@username = username
  end

  def self.userName
    @@username
  end

  def self.password=(password)
    @@password = password
  end

  def self.apiKey=(apiKey)
    @@apiKey = apiKey
  end

  def self.apiSecret=(apiSecret)
    @@apiSecret = apiSecret
  end

  def self.debug
    @@debug
  end

  def self.debug=(value)
    @@debug = value
  end

  def self.generateAuthHeaders( request, params=nil, apiKey=nil, apiSecret=nil, apiUser=nil, apiPass=nil )
    headers = {}
    apiKey = apiKey || @@apiKey
    apiSecret = apiSecret || @@apiSecret
    apiUser= apiUser || @@username
    apiPass = apiPass || @@password

    timestamp = Time.now.to_i

    unless request
      # Implied auth interceptor
      text_to_sign = "%s\n%s\n" % [apiKey,timestamp.to_s]
      unless apiUser.nil? or apiPass.nil?
        text_to_sign += "%s\n%s\n" % [apiUser,apiPass]
      end
    else
      # Authenticated api call
      method = request.method

      if @@debug
        print("Request Params: \n--------------\n")
        params.keys.each do |k|
          print("%s : %s\n" % [k, params[k]])
        end
      end

      # Strip query string from path, which is really a uri
      uri = URI::parse(request.path)
      path = uri.path
      query = uri.query.nil? ? "" : uri.query
      body_hash = ""

      if ['PUT', 'POST'].include? request.method
        payload = params[:payload]
        case payload
        when Message
          payload = payload.json_payload.to_json
        when Hash
          print "Did you mean to convert this Hash to json?\n"
          raise "API does not support x-www-form-urlencoded"
        when String
          payload = payload
          #headers['content_type'] = 'application/json'
        else
          raise "Unrecognized input to RestClient"
        end
        body_hash = OpenSSL::Digest::SHA256.hexdigest(payload)
        if @@debug
          print "PAYLOAD=%s\n" % payload
        end
      end
      text_to_sign = "%s\n%s\n%s\n%s\n%s\n" % [@@authCode,method,path,query,body_hash]
    end

    signature = OpenSSL::HMAC.hexdigest('sha256',apiSecret,text_to_sign)

    if @@debug
      print "text_to_sign: %s\n" % text_to_sign
      print "signature: %s\n" % signature
      print "apiKey: %s\n" % apiKey
      print "apiSecret: %s\n" % apiSecret
      print "timestamp: %s\n" % timestamp
    end

    unless request
      headers['token'] = apiKey
      headers['date'] = timestamp
      headers['signature'] = signature
    else
      headers['Cookie'] = "signature=%s:%s" % [@@authCode, signature]
    end

    headers
  end
  
  def self.injectAuthHeaders( request, params )

    if @@authCode.nil?  # or authCode expired
      self.login(@@username,@@password)
    end
    
    auth_headers = generateAuthHeaders( request, params, @@apiKey, @@apiSecret)

    # Add headers to request
    auth_headers.keys.each do |key|
      request[key] = auth_headers[key]
      if @@debug
        print "%s : %s\n" % [key, auth_headers[key]]
      end
    end

  end

  def self.login(user, pass)
    # Login

    data = generateAuthHeaders(nil)
    data[:user] = user
    data[:pass] = pass
    json = data.to_json

    if @@debug
      print "Request: %s\n" % json
    end

    @@in_login = true
    response = RestClient.post "#{@@baseURI}/auth", json, :content_type => :json, :accept => :json
    @@in_login = false
    hh = JSON.parse( response.body )
    if @@debug
      print "Response: %s\n" % response
    end
    @@authCode = hh['auth']
  end

  def self.logout
    # Invalidate creds
    @@authCode = nil
  end

  def self.harvestNextAuthCode(response)
    hh = JSON.parse(response)
    @@authCode = hh['auth']
  end

  def self.sign(user=nil,pass=nil)
    unless user.nil?
      @@username = user
      if pass.nil?
        require "io/console"
        print("Enter password for API user %s: " % user)
        pass = STDIN.noecho(&:gets).chomp
        print("\n")
      end
      @@password = pass unless pass.nil?
    end

    self.disableAutoAuth
    self.enableAutoAuth
    result = yield
    self.disableAutoAuth
    # Invalidate authCode
    @@authCode = nil
    #self.harvestNextAuthCode(result)
    return result
  end

  def self.disableAutoAuth
    @@auto_auth = false
    RestClient.reset_before_execution_procs
  end

  def self.enableAutoAuth
    @@auto_auth = true
    RestClient.add_before_execution_proc do |request, params|
      if @@auto_auth
        # guard against infinite recursion     
        #print "In_login = #{@@in_login}\n"
        #if request['signed']
        if true
          injectAuthHeaders(request, params) unless @@in_login
        else
          print "Not signing request\n"
        end
      end
    end
  end

  def self.baseURI=(url)
    @@baseURI = url
  end

  def self.URI(path)
    "%s%s" % [@@baseURI,path]
  end
end

end
